#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re
from pathlib import Path

from . import utils
from .column import *
from .configurable import Configurable
from .includes import *
from .metavars import *
from .schemas import *
from .version import *
from .writer import *


class HeaderFile(Configurable):
    __schema = Schema(
        {
            Optional(r'banner', default=''): Stripped(str),
            Optional(r'brief', default=''): Stripped(str),
            Optional(r'combined', default=True): bool,
            Optional(r'footer', default=''): Stripped(str),
            Optional(r'header', default=''): Stripped(str),
            Optional(r'includes', default=dict): {object: object},
            Optional(r'prologue', default=''): Stripped(str),
        }
    )

    def __init__(self, config, structs, vals):
        super().__init__(config)
        self.__dict__.update(HeaderFile.__schema.validate(vals))

        self.path = self.config.path.with_suffix('.hpp')

        self.structs = tuple(utils.coerce_collection(structs))
        assert len(self.structs) >= 1

        self.additional_header_files = []
        if not self.combined:
            self.additional_header_files = [HeaderFile(config, struct, vals) for struct in self.structs[1:]]
            self.structs = tuple(self.structs[:1])
            self.path = Path(self.config.path.parent, rf'{self.structs[0].name}.hpp')

        self.meta = MetaVars()
        for prefix in (r'file_', r'file::'):
            self.meta.push(rf'{prefix}name', self.path.name)
            self.meta.push(rf'{prefix}path', str(self.path))

        # includes
        with SchemaContext('includes'):
            self.includes = Includes(config, self.includes)

            # add any includes requested by member structs
            self.includes.internal = set(self.includes.internal)
            self.includes.external = set(self.includes.external)
            for struct in self.structs:
                for h in struct.includes.internal:
                    self.includes.internal.add(h)
                for h in struct.includes.external:
                    self.includes.external.add(h)
            self.includes.internal = sorted(self.includes.internal)
            self.includes.external = sorted(self.includes.external)

        # documentation
        if not self.brief and self.structs:
            self.brief = ', '.join([n.qualified_type for n in self.structs])
            self.brief = rf'Contains the definition{"s" if len(self.structs) > 1 else ""} of {self.brief}.'

        # banner
        if self.banner:
            self.banner = [s.rstrip() for s in self.banner.split('\n')]
            self.banner = [re.sub(r'(^|\n)// ?', r'\1', s) for s in self.banner]
            self.banner = [rf'// {s}'.rstrip() for s in self.banner]
            self.banner = '\n'.join(self.banner)

    def write(self, o: Writer):
        with MetaScope(self):
            if self.banner:
                o(rf'//{"-"*(120 - o.indent_width - 2)}')
                o(self.banner)

            o(rf'//{"-"*(120 - o.indent_width - 2)}')
            o(rf'// This file was generated by soagen v{VERSION_STRING} - do not modify it directly')
            o(rf'// https://marzer.github.io/soagen')
            o(rf'//{"-"*(120 - o.indent_width - 2)}')
            o(rf'#pragma once')
            o()

            if o.doxygen:
                o(
                    rf'''
                /// @file
                /// @brief {self.brief}
                ///
                /// @note The code and documentation in this file were generated by soagen - https://marzer.github.io/soagen
                '''
                )

            # internal includes (#include "blah.h")
            if self.includes.internal:
                o()
                for inc in self.includes.internal:
                    o(rf'#include "{inc}"')

            # stick the prologue here so users can use it to override the various
            # SOAGEN_ macros etc before including soagen.hpp
            o(
                rf'''
            {self.prologue}
            '''
            )

            # soagen.hpp
            o(
                rf'''
            #include <soagen.hpp>
            #if SOAGEN_VERSION_MAJOR != {VERSION[0]} || SOAGEN_VERSION_MINOR < {VERSION[1]}
                #error soagen version mismatch - expected v{VERSION[0]}.{VERSION[1]}.X
            #endif
            '''
            )

            # external/system includes (#include <blah.h>)
            # (deferred until later so we can take advantage of detect_includes() for the whole file)
            o('\n\n// __SOAGEN_EXTERNAL_HEADERS\n\n')

            # misc preprocessor boilerplate
            o(
                rf'''
            SOAGEN_PUSH_WARNINGS;
            SOAGEN_DISABLE_SPAM_WARNINGS;
            #if SOAGEN_CLANG >= 16
                #pragma clang diagnostic ignored "-Wunsafe-buffer-usage"
            #endif
            #if SOAGEN_MSVC
                #pragma inline_recursion(on)
            #endif
            #if SOAGEN_MSVC_LIKE
                #pragma push_macro("min")
                #pragma push_macro("max")
                #undef min
                #undef max
            #endif
            '''
            )

            # doxygen crap
            if o.doxygen:
                o(
                    rf'''
                //{"-"*(120 - o.indent_width - 2)}
                // doxygen safeguards
                //{"-"*(120 - o.indent_width - 2)}

                #if defined(DOXYGEN) || defined(__DOXYGEN) || defined(__DOXYGEN__) \
                    || defined(__doxygen__) || defined(__POXY__) || defined(__poxy__)

                #ifndef SOAGEN_DOXYGEN
                    #define SOAGEN_DOXYGEN 1
                #endif
                #ifndef SOAGEN_MAKE_NAME
                    #define SOAGEN_MAKE_NAME(...) static_assert(true)
                #endif
                #ifndef SOAGEN_MAKE_COLUMN
                    #define SOAGEN_MAKE_COLUMN(...) static_assert(true)
                #endif
                #ifndef SOAGEN_NODISCARD
                    #define SOAGEN_NODISCARD
                #endif
                #ifndef SOAGEN_NODISCARD_CTOR
                    #define SOAGEN_NODISCARD_CTOR
                #endif
                #ifndef SOAGEN_INLINE_GETTER
                    #define SOAGEN_INLINE_GETTER inline
                #endif
                #ifndef SOAGEN_PURE_GETTER
                    #define SOAGEN_PURE_GETTER
                #endif
                #ifndef SOAGEN_PURE_INLINE_GETTER
                    #define SOAGEN_PURE_INLINE_GETTER inline
                #endif
                #ifndef SOAGEN_ALWAYS_INLINE
                    #define SOAGEN_ALWAYS_INLINE inline
                #endif
                #ifndef SOAGEN_CPP20_CONSTEXPR
                    #define SOAGEN_CPP20_CONSTEXPR constexpr
                #endif
                #ifndef SOAGEN_HIDDEN
                    #define SOAGEN_HIDDEN(...)
                #endif
                #ifndef SOAGEN_HIDDEN_BASE
                    #define SOAGEN_HIDDEN_BASE(...)
                #endif
                #ifndef SOAGEN_HIDDEN_CONSTRAINT
                    #define SOAGEN_HIDDEN_CONSTRAINT(...)
                #endif
                #ifndef SOAGEN_HIDDEN_PARAM
                    #define SOAGEN_HIDDEN_PARAM(...)
                #endif
                #ifndef SOAGEN_ENABLE_IF_T
                    #define SOAGEN_ENABLE_IF_T(T, ...) T
                #endif
                #ifndef SOAGEN_ENABLE_IF
                    #define SOAGEN_ENABLE_IF(...)
                #endif
                #ifndef SOAGEN_REQUIRES
                    #define SOAGEN_REQUIRES(...)
                #endif
                #ifndef SOAGEN_EMPTY_BASES
                    #define SOAGEN_EMPTY_BASES
                #endif
                #ifndef SOAGEN_COLUMN
                    #define SOAGEN_COLUMN(...)
                #endif
                #ifndef SOAGEN_ALIGNED_COLUMN
                    #define SOAGEN_ALIGNED_COLUMN(...)
                #endif
                #if !defined(POXY_IMPLEMENTATION_DETAIL) && !(defined(__POXY__) || defined(__poxy__))
                    #define POXY_IMPLEMENTATION_DETAIL(...) __VA_ARGS__
                #endif

                #endif // doxygen

                '''
                )

            # forward declarations
            o(
                rf'''
            //{"-"*(120 - o.indent_width - 2)}
            // forward declarations + soagen internal boilerplate
            //{"-"*(120 - o.indent_width - 2)}
            '''
            )
            with HiddenFromDoxygen(o), ClangFormatOff(o):
                with Namespace(o, self.config.namespace):
                    for struct in self.structs:
                        struct.write_forward_declarations(o)
                o()
                with Namespace(o, 'soagen'):
                    for struct in self.structs:
                        struct.write_soagen_specializations(o)
                o()
                with Namespace(o, 'soagen::detail'):
                    names = set()
                    for struct in self.structs:
                        for col in struct.columns:
                            names.add(col.name)
                    names = sorted(list(names))
                    for name in names:
                        sanitized_name = name.replace('::', '_')
                        pp_define = rf'SOAGEN_NAME_{sanitized_name}'
                        o(
                            rf'''
                        #ifndef {pp_define}
                            #define {pp_define}
                            SOAGEN_MAKE_NAME({name});
                        #endif
                        '''
                        )
                    for struct in self.structs:
                        struct.write_soagen_detail_specializations(o)

            # header
            if self.header:
                o(
                    rf'''
                //{"-"*(120 - o.indent_width - 2)}
                // header
                //{"-"*(120 - o.indent_width - 2)}

                '''
                )
                o(self.header)

            # structs (class definitions, then outline member implementations)
            for struct in self.structs:
                o(
                    rf'''
                //{"-"*(120 - o.indent_width - 2)}
                // {struct.name}
                //{"-"*(120 - o.indent_width - 2)}
                '''
                )
                with Namespace(o, self.config.namespace):
                    struct.write_class_definition(o)
                    with HiddenFromDoxygen(o):
                        struct.write_outline_member_implementations(o)

            # footer
            if self.footer:
                o(
                    rf'''
                //{"-"*(120 - o.indent_width - 2)}
                // footer
                //{"-"*(120 - o.indent_width - 2)}

                {self.footer}
                '''
                )

            # clean up preprocessor boilerplate
            o(
                rf'''
            #if SOAGEN_MSVC_LIKE
                #pragma pop_macro("min")
                #pragma pop_macro("max")
            #endif
            #if SOAGEN_MSVC
                #pragma inline_recursion(off)
            #endif
            SOAGEN_POP_WARNINGS;'''
            )


__all__ = [r'HeaderFile']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT


class ConfigBase(object):
    pass


class Configurable(object):
    def __init__(self, cfg):
        assert cfg is not None

        if isinstance(cfg, ConfigBase):
            self.__cfg = cfg
            return

        try:
            c = cfg.config()
            if isinstance(c, ConfigBase):
                self.__cfg = c
                return
        except:
            pass

        try:
            c = cfg.config
            if isinstance(c, ConfigBase):
                self.__cfg = c
                return
        except:
            pass

        self.__cfg = None

    @property
    def config(self) -> ConfigBase:
        return self.__cfg



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

from . import utils
from .configurable import Configurable
from .schemas import *
from .variable import StaticVariable


class StructInjector(Configurable):
    __schema = Schema(
        {
            Optional(r'annotations', default=[]): And(
                ValueOrArray(str, name=r'annotations'),
                Use(lambda x: utils.remove_duplicates([s.strip() for s in x if s.strip()])),
            ),
            Optional(r'attributes', default=[]): And(
                ValueOrArray(str, name=r'attributes'),
                Use(lambda x: utils.remove_duplicates([s.strip() for s in x if s.strip()])),
            ),
            Optional(r'epilogue', default=''): Stripped(str),
            Optional(r'footer', default=''): Stripped(str),
            Optional(r'header', default=''): Stripped(str),
            Optional(r'prologue', default=''): Stripped(str),
            Optional(r'static_variables', default=list): [object],
        }
    )

    def __init__(self, config, vals):
        super().__init__(config)
        vals = StructInjector.__schema.validate(vals)
        self.__dict__.update(vals)

        self.static_variables = [StaticVariable(self.config, v) for v in self.static_variables]


__all__ = [r'StructInjector']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re
from pathlib import Path

from . import log, utils
from .configurable import ConfigBase
from .errors import *
from .header_file import HeaderFile
from .injectors import StructInjector
from .metavars import *
from .natvis_file import NatvisFile
from .schemas import *
from .struct import Struct
from .type_list import *

try:
    import pytomlpp as toml  # fast; based on toml++ (C++)
except ImportError:
    try:
        import tomllib as toml  # PEP 680
    except ImportError:
        import tomli as toml


class Config(ConfigBase):
    __schema = Schema(
        {
            Optional(r'hpp', default=dict): {object: object},
            Optional(r'namespace', default=''): And(
                str,
                Use(lambda x: x.strip(': \t\n\f\b\v').split('::')),
                Use(lambda x: [s.strip(': \t\n\f\b\v') for s in x]),
                Use(lambda x: '::'.join(x).replace('::::', '::')),
            ),
            Optional(r'structs', default=dict): {
                Stripped(str, allow_empty=False, name=r'struct name'): {object: object}
            },
            Optional(r'allocator', default=r'soagen::allocator'): Stripped(str),
            Optional(r'all_structs', default=dict): {object: object},
        }
    )

    def __init__(self, path):
        assert isinstance(path, Path)
        self.path = path.resolve()
        self.dir = path.parent

        self.meta_stack = MetaStack()
        self.meta = MetaVars()
        self.meta.push('tab', '\t')
        self.meta.push('space', ' ')
        self.meta_stack.push(self.meta)

        with SchemaContext(self.path.name):
            try:
                cfg = toml.loads(utils.read_all_text_from_file(self.path, logger=log.i))
                self.__dict__.update(Config.__schema.validate(cfg))
            except Exception as ex:
                raise SchemaError(str(ex), None)

            # namespace
            if self.namespace:
                if self.namespace in (r'std', r'soagen', r'soagen::detail') or self.namespace.startswith(r'std::'):
                    raise SchemaError(rf"namespace: '{self.namespace}' is reserved", None)
                self.meta.push('namespace', self.namespace)
                self.meta.push('namespace::name', self.namespace)
                self.meta.push('namespace::start', f'namespace {self.namespace}\n{{')
                self.meta.push('namespace::end', r'}')
                self.meta.push('namespace::scope', rf'{self.namespace}::')
            else:
                log.w(rf'{current_schema_context()}namespace: not set! polluting the global namespace is bad practice')
                self.meta.push('namespace', '')
                self.meta.push('namespace::name', '')
                self.meta.push('namespace::start', '')
                self.meta.push('namespace::end', '')
                self.meta.push('namespace::scope', '')

            # injectors for the 'all_X' sections
            self.all_structs = StructInjector(self, self.all_structs)

            # structs
            self.structs = [(k, v) for k, v in self.structs.items()]
            for i in range(len(self.structs)):
                with SchemaContext(rf"struct '{self.structs[i][0]}'"):
                    self.structs[i] = Struct(self, self.structs[i][0], self.structs[i][1])
            self.structs = tuple(sorted(self.structs, key=lambda s: s.name))
            self.struct_types = TypeList([s.type for s in self.structs])
            self.meta.push('struct_names', ', '.join([s.name for s in self.structs]))
            self.meta.push('struct_types', ', '.join([s.type for s in self.structs]))
            self.meta.push('qualified_struct_names', ', '.join([s.qualified_type for s in self.structs]))
            self.meta.push('qualified_struct_types', ', '.join([s.qualified_type for s in self.structs]))
            index = 0
            for struct in self.structs:
                struct.set_index(index)
                index += 1

            # hpp files (needs a little bit of special handling because of the 'combined' option)
            self.hpp = [HeaderFile(self, self.structs, self.hpp)]
            self.hpp += [*self.hpp[0].additional_header_files]
            self.hpp[0].additional_header_files = []

            # all outputs
            self.natvis = NatvisFile(self, self.structs)
            self.all_outputs = [self.natvis, *self.hpp]
            log.d('\n  -> '.join([str(self.path)] + [str(o.path) for o in self.all_outputs]))
            for o in self.all_outputs:
                if o.path.is_dir():
                    raise Error(rf"invalid output '{o.path}': outputs cannot be existing directories")
                if o.path == self.path:
                    raise Error(rf"invalid output '{o.path}': outputs cannot overwrite input")
                if o.path.name.lower() == r'soagen.hpp':
                    raise Error(rf"invalid output '{o.path}': 'soagen.hpp' is a reserved name")


__all__ = [r'Config']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re

from . import cpp, utils
from .column import Column
from .configurable import *
from .schemas import *
from .writer import *


class Variable(Configurable):
    __schema = Schema(
        {
            r'name': Stripped(str, allow_empty=False, name=r'variable name'),
            r'type': Stripped(str, allow_empty=False, name=r'variable type'),
            Optional(r'param_type', default=''): Stripped(str),
            Optional(r'default', default=None): And(
                Or(str, float, int), Use(lambda x: x.strip() if isinstance(x, str) else str(x))
            ),
            Optional(r'alignment', default=0): And(
                Or(int, str),
                Use(int),
                lambda x: x <= 0 or utils.is_pow2(x),
                error=r'alignment must be a power-of-two integer',
            ),
        }
    )

    def __init__(self, struct, vals):
        super().__init__(struct)
        self.struct = struct
        self.index = -1  # set by the struct

        vals = Variable.__schema.validate(vals)
        self.__dict__.update(vals)

        valid = cpp.is_valid_identifier(self.name)
        if not valid[0]:
            raise SchemaError(rf"name: '{self.name}': {valid[1]}", None)

        if self.default is None:
            self.default = ''

        self.pointer_type = rf'{self.type}*'
        if re.fullmatch(r'[a-zA-Z_][a-zA-Z_0-9:]*', self.type):
            self.const_pointer_type = rf'const {self.pointer_type}'
        else:
            self.const_pointer_type = rf'std::add_const_t<{self.type}>*'

        self.columns = [Column(self)]


class StaticVariable(Configurable):
    __schema = Schema(
        {
            r'name': Stripped(str, allow_empty=False, name=r'static variable name'),
            r'type': Stripped(str, allow_empty=False, name=r'static variable type'),
            r'value': And(Or(str, float, int), Use(lambda x: x.strip() if isinstance(x, str) else str(x))),
            Optional(r'const', default=None): Or('', r'const', r'constexpr', bool),
            Optional(r'access', default=r'public'): Or('public', r'protected', r'private'),
            Optional(r'brief', default=''): Stripped(str),
        }
    )

    def __init__(self, cfg, vals):
        super().__init__(cfg)
        self.__dict__.update(StaticVariable.__schema.validate(vals))

        valid = cpp.is_valid_identifier(self.name)
        if not valid[0]:
            raise SchemaError(rf"name: '{self.name}': {valid[1]}", None)

        if not self.value:
            raise SchemaError(rf"value: cannot be blank", None)

        if self.const is None:
            self.const = 'constexpr'
        elif isinstance(self.const, bool):
            self.const = 'const' if self.const else ''

    def write(self, o: Writer):
        if self.brief:
            o(f'/// @brief {self.brief}')
        s = 'static '
        if self.const != 'constexpr' and self.value is not None:
            s += 'inline '
        if self.const != 'const':
            s += f'{self.const} '
        s += f'{self.type} '
        if self.const == 'const':
            s += f'{self.const} '
        s += self.name
        if self.value is not None:
            s += rf' = {self.value}'

        s += ';'
        o(s)


__all__ = [r'Variable', r'StaticVariable']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

"""
Misc utility functions.
"""

import shutil
import subprocess
from pathlib import Path

from misk import *

from . import paths


def clang_format(s, cwd=None):
    cwd = Path.cwd() if cwd is None else cwd
    # respect the user's .clang-format file if one exists, but
    # fallback to the internal one (so there's always a sane default)
    style_dir = cwd
    style_file = None
    while True:
        file = Path(style_dir, r'.clang-format')
        if file.is_file():
            style_file = file
            break
        dir = style_dir.parent
        if not dir or dir == style_dir or not dir.is_dir():
            break
        style_dir = dir
    if style_file is None:
        style_file = Path(paths.HPP, r'.clang-format')
        assert style_file.is_file()

    result = subprocess.run(
        [r'clang-format', rf'--style=file:{style_file}'],
        capture_output=True,
        cwd=str(cwd),
        encoding='utf-8',
        check=True,
        input=s,
    )
    return str(result.stdout)


def is_tool(name):
    return shutil.which(name) is not None


def make_divider(text: str, text_col=40, pattern='-', line_length=120) -> str:
    text = "//{}  {}  ".format(repeat_pattern(pattern, text_col - 2), text)
    if len(text) < line_length:
        return text + repeat_pattern(pattern, line_length - len(text))
    else:
        return text



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import re
from pathlib import Path

from . import utils


def get_relative_path(path: Path, relative_to: Path) -> Path:
    relative_root = relative_to
    relative = None
    relative_prefix = ''
    while True:
        try:
            relative = path.relative_to(relative_root)
            break
        except:
            if not relative_root.is_dir() or relative_root.parent == relative_root:
                break
            relative_prefix = rf'../{relative_prefix}'
            relative_root = relative_root.parent
            pass
    if relative is not None:
        return Path(rf'{relative_prefix}{relative}')
    return path


class Preprocessor:

    """
    A simple C++ file 'preprocessor' for almalgamating source files.
    """

    __re_includes = re.compile(r'^\s*#\s*include\s+"(.+?)".*?$', re.I | re.M)
    __re_pragma_once = re.compile(r'^\s*#\s*pragma\s+once\s*$', re.M)
    __repeatable_headers = (r'header_start.hpp', r'header_end.hpp')

    def __init__(self, file):
        self.__processed_files = set()
        self.__once_only = set()
        self.__directory_stack = [Path.cwd()]
        self.__include_stack = []
        self.__entry_root = None
        self.__string = self.__preprocess(file)
        # do some cleanup of the preprocessed file
        while True:
            s = self.__string
            # trailing whitespace
            s = re.sub('([^ \t])[ \t]+\n', r'\1\n', s)
            # double blank lines
            s = re.sub('\n\n\n+', '\n\n', s)
            # magic comments
            s = re.sub(rf'(?://(?:[#!</]|[ \t]?(?:\^\^\^|vvv))).*?($)', r'\1', s, flags=re.MULTILINE)
            # consecutive header separators
            header_separator = r'(?://\*\*\*\**[ \t]+[a-zA-Z0-9_/.-]+[ \t]+\*\*\*\*+\n)'
            s = re.sub(rf'(?:{header_separator}(?:[ \t]*\n)*)+({header_separator})', r'\1', s)
            # weird spacing edge case around pp directives
            s = re.sub('\n[}]\n#', r'\n}\n\n#', s)
            # blank lines following a comma, opening bracket, or colon
            s = re.sub(r'([{(\[:,])\n\n', r'\1\n', s)
            # blank lines preceeding closing brackets
            s = re.sub(r'\n\n([ \t]*[})])', r'\n\1', s)
            # empty #if blocks
            s = re.sub(r'#if(n?def)?[ \t]+[a-zA-Z0-9_]+\s*?\n+\s*?#endif', r'\n', s)
            # empty #else blocks
            s = re.sub(r'#else\s*?\n+\s*?#endif', r'#endif\n', s)
            # #if SOAGEN_DOXYGEN ... #else ... #endif
            s = re.sub(r'#if[ \t]+SOAGEN_DOXYGEN\s*?(?:\n+[^#]*?)?\n+\s*?#else\s*?\n+([^#]*?)\n+\s*?#endif', r'\1\n', s)
            # #if !SOAGEN_DOXYGEN ... #else ... #endif
            s = re.sub(
                r'#if[ \t]+![ \t]*SOAGEN_DOXYGEN\s*?\n+([^#]*?)\n+\s*?#else\s*?(?:\n+[^#]*?)?\n+\s*?#endif', r'\1\n', s
            )
            # #if SOAGEN_DOXYGEN ... #endif
            s = re.sub(r'#if[ \t]+SOAGEN_DOXYGEN\s*?(?:\n+[^#]*?)?\n+\s*?#endif', r'\n', s)
            # #if !SOAGEN_DOXYGEN ... #endif
            s = re.sub(r'#if[ \t]+![ \t]*SOAGEN_DOXYGEN\s*?\n+([^#]*?)\n+\s*?#endif', r'\1\n', s)
            if s == self.__string:
                break
            self.__string = s
        self.__string = self.__string.strip() + '\n'

    def __preprocess(self, incl):
        if not isinstance(incl, (Path, str)):  # a regex match object
            incl = incl.group(1).strip()
        if isinstance(incl, str):
            incl = Path(incl.strip().replace('\\', r'/'))
        if not incl.is_absolute():
            incl = Path(self.__directory_stack[-1], incl).resolve()
        if not incl.is_file():
            sibling = Path(incl.parent.parent, 'hpp', incl.name)
            if sibling.is_file():
                incl = sibling
        self.__processed_files.add(incl)
        if incl in self.__once_only:
            return ''
        self.__include_stack.append(incl)

        text = utils.read_all_text_from_file(incl, logger=True).strip() + '\n'
        text = text.replace('\r\n', '\n')  # convert windows newlines

        self.__directory_stack.append(incl.parent)
        if self.__re_pragma_once.search(text):
            self.__once_only.add(incl)
        if len(self.__include_stack) == 1 and self.__entry_root is None:
            self.__entry_root = incl.parent.resolve()
        if len(self.__include_stack) > 1:
            text = self.__re_pragma_once.sub('', text)

        text = self.__re_includes.sub(lambda m: self.__preprocess(m), text, 0)

        # explicit 'strip this' blocks
        text = re.sub(r'(?:\n[ \t]*)?//[#!][ \t]*[{][{].*?//[#!][ \t]*[}][}].*?\n', '\n', text, flags=re.S)

        incl_name = incl.name
        incl = get_relative_path(incl, self.__entry_root)
        incl = str(incl).replace('\\', r'/').strip('/')

        if len(self.__include_stack) > 1 and incl_name not in self.__repeatable_headers:
            header = utils.make_divider(str(incl), 10, pattern=r'*')

            footer = ''
            if len(self.__include_stack) > 2:
                footer = get_relative_path(self.__include_stack[-2], self.__entry_root)
                footer = str(footer).replace('\\', r'/').strip('/')
                footer = utils.make_divider(footer, 10, pattern=r'*')

            text = f'\n\n{header}\n\n{text}\n\n{footer}'.rstrip()

        self.__include_stack.pop()
        self.__directory_stack.pop()
        return '\n\n' + text + '\n\n'

    def __str__(self):
        return self.__string

    def processed_files(self):
        out = list(self.__processed_files)
        out.sort()
        return out


__all__ = [r'Preprocessor']



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT


class TypeList(object):
    def __init__(self, *types):
        self.__types = []
        if types:
            for t in types:
                self.__flattening_insert(t)

    def __flattening_insert(self, typ):
        if isinstance(typ, TypeList):
            self.__flattening_insert(typ.__types)
        elif isinstance(typ, (list, tuple, set)):
            for t in typ:
                self.__flattening_insert(t)
        else:
            assert isinstance(typ, str)
            self.__types.append(typ)

    def __str__(self):
        return r', '.join(self.__types)

    def __repr__(self):
        return repr(self.__types)

    def __add__(lhs, rhs):
        assert isinstance(lhs, TypeList)
        assert isinstance(rhs, (TypeList, list, tuple, set, str))
        return TypeList(lhs, rhs)

    def __iadd__(lhs, rhs):
        assert isinstance(lhs, TypeList)
        assert isinstance(rhs, (TypeList, list, tuple, set, str))
        lhs.__flattening_insert(rhs)
        return lhs

    def __len__(self):
        return len(self.__types)

    def __bool__(self):
        return len(self.__types) > 0

    def __iter__(self):
        for typ in self.__types:
            yield typ

    def __getitem__(self, key):
        return self.__types[key]

    def __contains__(self, val):
        assert isinstance(val, str)
        return val in self.__types

    def add_prefix(self, prefix):
        assert isinstance(prefix, str)
        if prefix:
            for i in range(len(self.__types)):
                self.__types[i] = prefix + self.__types[i]
        return self

    def add_suffix(self, suffix):
        assert isinstance(suffix, str)
        if suffix:
            for i in range(len(self.__types)):
                self.__types[i] = self.__types[i] + suffix
        return self

    def sort(self):
        self.__types.sort()
        return self



#!/usr/bin/env python3
# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

"""
Logging functions.
"""

import logging
import sys

from colorama import Fore, Style
from colorama import init as colorama_init

_min_level = logging.INFO
_treat_warnings_as_errors = False
_on_warning = None
_on_error = None
_initialized = False
_styles_allowed = True

STYLE_RED = rf'{Fore.RED}{Style.BRIGHT}'
STYLE_GREEN = rf'{Fore.GREEN}{Style.BRIGHT}'
STYLE_BLUE = rf'{Fore.BLUE}{Style.BRIGHT}'
STYLE_CYAN = rf'{Fore.CYAN}{Style.BRIGHT}'
STYLE_YELLOW = rf'{Fore.YELLOW}{Style.BRIGHT}'
STYLE_MAGENTA = rf'{Fore.MAGENTA}{Style.BRIGHT}'

STYLE_RED_DIM = rf'{Fore.RED}{Style.DIM}'
STYLE_GREEN_DIM = rf'{Fore.GREEN}{Style.DIM}'
STYLE_BLUE_DIM = rf'{Fore.BLUE}{Style.DIM}'
STYLE_CYAN_DIM = rf'{Fore.CYAN}{Style.DIM}'
STYLE_YELLOW_DIM = rf'{Fore.YELLOW}{Style.DIM}'
STYLE_MAGENTA_DIM = rf'{Fore.MAGENTA}{Style.DIM}'

STYLE_RESET = rf'{Style.RESET_ALL}'


def _init(min_level: int, treat_warnings_as_errors: bool, on_warning, on_error, allow_styles=True, allow_reinit=False):
    global _initialized
    if _initialized and not bool(allow_reinit):
        return

    global _min_level
    _min_level = int(min_level)

    global _treat_warnings_as_errors
    _treat_warnings_as_errors = bool(treat_warnings_as_errors)

    global _on_warning
    _on_warning = on_warning

    global _on_error
    _on_error = on_error

    if not _initialized:
        global _styles_allowed
        _styles_allowed = bool(allow_styles)
        colorama_init(strip=not _styles_allowed)

    _initialized = True


def init(min_level=logging.INFO, treat_warnings_as_errors=False, on_warning=None, on_error=None, allow_styles=True):
    _init(
        min_level=min_level,
        treat_warnings_as_errors=treat_warnings_as_errors,
        on_warning=on_warning,
        on_error=on_error,
        allow_styles=allow_styles,
        allow_reinit=False,
    )


def reinit(min_level=logging.INFO, treat_warnings_as_errors=False, on_warning=None, on_error=None):
    _init(
        min_level=min_level,
        treat_warnings_as_errors=treat_warnings_as_errors,
        on_warning=on_warning,
        on_error=on_error,
        allow_reinit=True,
    )


def styles_allowed():
    global _styles_allowed
    return _styles_allowed


def clear_hooks():
    global _on_warning
    _on_warning = None

    global _on_error
    _on_error = None


def d(*message):
    if _min_level > logging.DEBUG:
        return
    print(*message, file=sys.stdout)


def i(*message):
    if _min_level > logging.INFO:
        return
    print(*message, file=sys.stdout)


def w(*message):
    if _min_level > logging.WARNING:
        return
    if _treat_warnings_as_errors:
        print(
            rf'{STYLE_RED}error:{STYLE_RESET}',
            *message,
            rf'{STYLE_YELLOW}(warning treated as error){STYLE_RESET}',
            file=sys.stderr,
        )
        if _on_error is not None:
            _on_error()
    else:
        print(rf'{STYLE_YELLOW}warning:{STYLE_RESET}', *message, file=sys.stderr)
        if _on_warning is not None:
            _on_warning()


def e(*message):
    if _min_level > logging.ERROR:
        return
    print(rf'{STYLE_RED}error:{STYLE_RESET}', *message, file=sys.stderr)
    if _on_error is not None:
        _on_error()


__all__ = [
    'init',
    'reinit',
    'clear_hooks',
    'styles_allowed',
    'd',
    'i',
    'w',
    'e',
    r'STYLE_RED',
    r'STYLE_GREEN',
    r'STYLE_BLUE',
    r'STYLE_CYAN',
    r'STYLE_YELLOW',
    r'STYLE_MAGENTA',
    r'STYLE_RED_DIM',
    r'STYLE_GREEN_DIM',
    r'STYLE_BLUE_DIM',
    r'STYLE_CYAN_DIM',
    r'STYLE_YELLOW_DIM',
    r'STYLE_MAGENTA_DIM',
    r'STYLE_RESET',
]
